# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The `blazingmq.testing` module provides testing utilities.

Nothing in this module should ever be used outside of the context of performing
integration tests on your subscribers and publishers.
"""

from ._monitors import BasicHealthMonitor

HostHealth = BasicHealthMonitor
"""This is an alias of `.BasicHealthMonitor`.

This was originally exposed only for testing purposes, but is now supported for
general use. The original name has been retained for backwards compatibility,
but new code should prefer to use `blazingmq.BasicHealthMonitor` directly.
"""
